import { __awaiter } from 'tslib';
import * as i0 from '@angular/core';
import { Injectable, InjectionToken, NgModule } from '@angular/core';
import { HttpResponse } from '@angular/common/http';
import { fromEvent, merge, defer, from, throwError, Subject } from 'rxjs';
import { map, share, catchError } from 'rxjs/operators';
import * as i1 from '@ionic/storage-angular';
import { Storage, IonicStorageModule } from '@ionic/storage-angular';

class CacheStorageService {
    constructor(storage, keyPrefix) {
        this.storage = storage;
        this.keyPrefix = keyPrefix;
    }
    create() {
        return this.storage.create();
    }
    set(key, value) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.storage.set(this.buildKey(key), value);
        });
    }
    remove(key) {
        return __awaiter(this, void 0, void 0, function* () {
            return this.storage.remove(this.buildKey(key));
        });
    }
    get(key) {
        return __awaiter(this, void 0, void 0, function* () {
            let value = yield this.storage.get(this.buildKey(key));
            return !!value ? Object.assign({ key: key }, value) : null;
        });
    }
    exists(key) {
        return __awaiter(this, void 0, void 0, function* () {
            return !!(yield this.storage.get(this.buildKey(key)));
        });
    }
    all() {
        return __awaiter(this, void 0, void 0, function* () {
            let items = [];
            yield this.storage.forEach((val, key) => {
                if (this.isCachedItem(key, val)) {
                    items.push(Object.assign({ key: this.debuildKey(key) }, val));
                }
            });
            return items;
        });
    }
    /**
     * @description Returns whether or not an object is a cached item.
     * @return {boolean}
     */
    isCachedItem(key, item) {
        return (item && item.expires && item.type && key.startsWith(this.keyPrefix));
    }
    /**
     * Makes sure that the key is prefixed properly
     * @return {string}
     */
    buildKey(key) {
        if (key.startsWith(this.keyPrefix)) {
            return key;
        }
        return this.keyPrefix + key;
    }
    /**
     * Makes sure that the key isn't prefixed
     * @return {string}
     */
    debuildKey(key) {
        if (key.startsWith(this.keyPrefix)) {
            return key.substr(this.keyPrefix.length);
        }
        return key;
    }
}

const MESSAGES = {
    0: 'Cache initialization error: ',
    1: 'Cache is not enabled.',
    2: 'Cache entry already expired: ',
    3: 'No such key: ',
    4: 'No entries were deleted, because browser is offline.',
};
/**
 * @description Check if it's an HttpResponse
 * @param {any} data - Variable to test
 * @return {boolean} - data from cache
 */
const isHttpResponse = (data) => {
    let orCondition = data &&
        typeof data === 'object' &&
        data.hasOwnProperty('status') &&
        data.hasOwnProperty('statusText') &&
        data.hasOwnProperty('headers') &&
        data.hasOwnProperty('url') &&
        data.hasOwnProperty('body');
    return data && (data instanceof HttpResponse || orCondition);
};
const isJsOrResponseType = (data) => {
    const jsType = data.type === 'undefined' ||
        data.type === 'object' ||
        data.type === 'boolean' ||
        data.type === 'number' ||
        data.type === 'bigint' ||
        data.type === 'string' ||
        data.type === 'symbol' ||
        data.type === 'function';
    const responseType = data.type === 'response';
    return responseType || jsType;
};
class CacheService {
    constructor(_storage) {
        this._storage = _storage;
        this.ttl = 60 * 60; // one hour
        this.cacheEnabled = true;
        this.invalidateOffline = false;
        this.networkStatus = true;
        this.watchNetworkInit();
        this.loadCache();
    }
    loadCache() {
        return __awaiter(this, void 0, void 0, function* () {
            try {
                yield this.create();
                this.cacheEnabled = true;
            }
            catch (e) {
                this.cacheEnabled = false;
                console.error(MESSAGES[0], e);
            }
        });
    }
    create() {
        return this._storage.create();
    }
    /**
     * @description Disable or enable cache
     */
    enableCache(enable = true) {
        this.cacheEnabled = enable;
    }
    /**
     * @description Delete DB table and create new one
     * @return {Promise<any>}
     */
    resetDatabase() {
        return __awaiter(this, void 0, void 0, function* () {
            yield this.create();
            let items = yield this._storage.all();
            return Promise.all(items.map((item) => this.removeItem(item.key)));
        });
    }
    /**
     * @description Set default TTL
     * @param {number} ttl - TTL in seconds
     */
    setDefaultTTL(ttl) {
        return (this.ttl = ttl);
    }
    /**
     * @description Set if expired cache should be invalidated if device is offline
     * @param {boolean} offlineInvalidate
     */
    setOfflineInvalidate(offlineInvalidate) {
        this.invalidateOffline = !offlineInvalidate;
    }
    /**
     * @description Start watching if devices is online or offline
     */
    watchNetworkInit() {
        this.networkStatus = navigator.onLine;
        const connect = fromEvent(window, 'online').pipe(map(() => true)), disconnect = fromEvent(window, 'offline').pipe(map(() => false));
        this.networkStatusChanges = merge(connect, disconnect).pipe(share());
        this.networkStatusChanges.subscribe((status) => {
            this.networkStatus = status;
        });
    }
    /**
     * @description Stream of network status changes
     * * @return {Observable<boolean>} network status stream
     */
    getNetworkStatusChanges() {
        return this.networkStatusChanges;
    }
    /**
     * @description Check if devices is online
     * @return {boolean} network status
     */
    isOnline() {
        return this.networkStatus;
    }
    /**
     * @description Save item to cache
     * @param {string} key - Unique key
     * @param {any} data - Data to store
     * @param {string} [groupKey] - group key
     * @param {number} [ttl] - TTL in seconds
     * @return {Promise<any>} - saved data
     */
    saveItem(key, data, groupKey = 'none', ttl = this.ttl) {
        if (!this.cacheEnabled) {
            throw new Error(MESSAGES[1]);
        }
        if (Blob.name === data.constructor.name) {
            return this.saveBlobItem(key, data, groupKey, ttl);
        }
        const expires = new Date().getTime() + ttl * 1000, type = isHttpResponse(data) ? 'response' : typeof data, value = JSON.stringify(data);
        return this._storage.set(key, {
            value,
            expires,
            type,
            groupKey,
        });
    }
    /**
     * @description Save blob item to cache
     * @param {string} key - Unique key
     * @param {any} blob - Blob to store
     * @param {string} [groupKey] - group key
     * @param {number} [ttl] - TTL in seconds
     * @return {Promise<any>} - saved data
     */
    saveBlobItem(key, blob, groupKey = 'none', ttl = this.ttl) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.cacheEnabled) {
                throw new Error(MESSAGES[1]);
            }
            const expires = new Date().getTime() + ttl * 1000, type = blob.type;
            try {
                const base64data = yield this.asBase64(blob);
                const value = JSON.stringify(base64data);
                return this._storage.set(key, {
                    value,
                    expires,
                    type,
                    groupKey,
                });
            }
            catch (error) {
                throw new Error(error);
            }
        });
    }
    // Technique derived from: https://stackoverflow.com/a/18650249
    asBase64(blob) {
        return new Promise((resolve, reject) => {
            const reader = new FileReader();
            reader.readAsDataURL(blob);
            reader.onloadend = () => {
                const base64data = reader.result;
                resolve(base64data);
            };
            reader.onerror = (event) => {
                reject(event);
                reader.abort();
            };
        });
    }
    /**
     * @description Delete item from cache
     * @param {string} key - Unique key
     * @return {Promise<any>} - query execution promise
     */
    removeItem(key) {
        if (!this.cacheEnabled) {
            throw new Error(MESSAGES[1]);
        }
        return this._storage.remove(key);
    }
    /**
     * @description Removes all items with a key that matches pattern
     * @return {Promise<any>}
     */
    removeItems(pattern) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.cacheEnabled) {
                throw new Error(MESSAGES[1]);
            }
            let regex = new RegExp(`^${pattern.split('*').join('.*')}$`);
            let items = yield this._storage.all();
            return Promise.all(items
                .map((item) => item.key)
                .filter((key) => key && regex.test(key))
                .map((key) => this.removeItem(key)));
        });
    }
    /**
     * @description Get item from cache without expire check etc.
     * @param {string} key - Unique key
     * @return {Promise<any>} - data from cache
     */
    getRawItem(key) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.cacheEnabled) {
                throw new Error(MESSAGES[1]);
            }
            try {
                let data = yield this._storage.get(key);
                if (!!data) {
                    return data;
                }
                throw new Error('');
            }
            catch (err) {
                throw new Error(MESSAGES[3] + key);
            }
        });
    }
    getRawItems() {
        return __awaiter(this, void 0, void 0, function* () {
            return this._storage.all();
        });
    }
    /**
     * @description Check if item exists in cache regardless if expired or not
     * @param {string} key - Unique key
     * @return {Promise<boolean | string>} - boolean - true if exists
     */
    itemExists(key) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.cacheEnabled) {
                throw new Error(MESSAGES[1]);
            }
            return this._storage.exists(key);
        });
    }
    /**
     * @description Get item from cache with expire check and correct type assign
     * @param {string} key - Unique key
     * @return {Promise<any>} - data from cache
     */
    getItem(key) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.cacheEnabled) {
                throw new Error(MESSAGES[1]);
            }
            let data = yield this.getRawItem(key);
            if (data.expires < new Date().getTime() &&
                (this.invalidateOffline || this.isOnline())) {
                throw new Error(MESSAGES[2] + key);
            }
            return CacheService.decodeRawData(data);
        });
    }
    getOrSetItem(key, factory, groupKey, ttl) {
        return __awaiter(this, void 0, void 0, function* () {
            let val;
            try {
                val = yield this.getItem(key);
            }
            catch (error) {
                val = yield factory();
                yield this.saveItem(key, val, groupKey, ttl);
            }
            return val;
        });
    }
    /**
     * @description Decode raw data from DB
     * @param {any} data - Data
     * @return {any} - decoded data
     */
    static decodeRawData(data) {
        return __awaiter(this, void 0, void 0, function* () {
            let dataJson = JSON.parse(data.value);
            if (isJsOrResponseType(data)) {
                if (isHttpResponse(dataJson)) {
                    let response = {
                        body: dataJson._body || dataJson.body,
                        status: dataJson.status,
                        headers: dataJson.headers,
                        statusText: dataJson.statusText,
                        url: dataJson.url,
                    };
                    return new HttpResponse(response);
                }
                return dataJson;
            }
            else {
                // Technique derived from: https://stackoverflow.com/a/36183085
                const response = yield fetch(dataJson);
                return response.blob();
            }
        });
    }
    /**
     * @description Load item from cache if it's in cache or load from origin observable
     * @param {string} key - Unique key
     * @param {any} observable - Observable with data
     * @param {string} [groupKey] - group key
     * @param {number} [ttl] - TTL in seconds
     * @return {Observable<any>} - data from cache or origin observable
     */
    loadFromObservable(key, observable, groupKey, ttl) {
        if (!this.cacheEnabled)
            return observable;
        observable = observable.pipe(share());
        return defer(() => {
            return from(this.getItem(key)).pipe(catchError((e) => {
                observable.subscribe((res) => {
                    return this.saveItem(key, res, groupKey, ttl);
                }, (error) => {
                    return throwError(error);
                });
                return observable;
            }));
        });
    }
    /**
     * @description Load item from cache if it's in cache or load from origin observable
     * @param {string} key - Unique key
     * @param {any} observable - Observable with data
     * @param {string} [groupKey] - group key
     * @param {number} [ttl] - TTL in seconds
     * @param {string} [delayType='expired']
     * @param {string} [metaKey] - property on T to which to assign meta data
     * @return {Observable<any>} - data from cache or origin observable
     */
    loadFromDelayedObservable(key, observable, groupKey, ttl = this.ttl, delayType = 'expired', metaKey) {
        if (!this.cacheEnabled)
            return observable;
        const observableSubject = new Subject();
        observable = observable.pipe(share());
        const subscribeOrigin = () => {
            observable.subscribe((res) => {
                this.saveItem(key, res, groupKey, ttl);
                observableSubject.next(res);
                observableSubject.complete();
            }, (err) => {
                observableSubject.error(err);
            }, () => {
                observableSubject.complete();
            });
        };
        this.getItem(key)
            .then((data) => {
            if (metaKey) {
                data[metaKey] = data[metaKey] || {};
                data[metaKey].fromCache = true;
            }
            observableSubject.next(data);
            if (delayType === 'all') {
                subscribeOrigin();
            }
            else {
                observableSubject.complete();
            }
        })
            .catch((e) => {
            this.getRawItem(key)
                .then((res) => __awaiter(this, void 0, void 0, function* () {
                let result = yield CacheService.decodeRawData(res);
                if (metaKey) {
                    result[metaKey] = result[metaKey] || {};
                    result[metaKey].fromCache = true;
                }
                observableSubject.next(result);
                subscribeOrigin();
            }))
                .catch(() => subscribeOrigin());
        });
        return observableSubject.asObservable();
    }
    /**
     * Perform complete cache clear
     * @return {Promise<any>}
     */
    clearAll() {
        if (!this.cacheEnabled) {
            throw new Error(MESSAGES[2]);
        }
        return this.resetDatabase();
    }
    /**
     * @description Remove all expired items from cache
     * @param {boolean} ignoreOnlineStatus -
     * @return {Promise<any>} - query promise
     */
    clearExpired(ignoreOnlineStatus = false) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.cacheEnabled) {
                throw new Error(MESSAGES[2]);
            }
            if (!this.isOnline() && !ignoreOnlineStatus) {
                throw new Error(MESSAGES[4]);
            }
            let items = yield this._storage.all();
            let datetime = new Date().getTime();
            return Promise.all(items
                .filter((item) => item.expires < datetime)
                .map((item) => this.removeItem(item.key)));
        });
    }
    /**
     * @description Remove all item with specified group
     * @param {string} groupKey - group key
     * @return {Promise<any>} - query promise
     */
    clearGroup(groupKey) {
        return __awaiter(this, void 0, void 0, function* () {
            if (!this.cacheEnabled) {
                throw new Error(MESSAGES[2]);
            }
            let items = yield this._storage.all();
            return Promise.all(items
                .filter((item) => item.groupKey === groupKey)
                .map((item) => this.removeItem(item.key)));
        });
    }
}
CacheService.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.1.4", ngImport: i0, type: CacheService, deps: [{ token: CacheStorageService }], target: i0.ɵɵFactoryTarget.Injectable });
CacheService.ɵprov = i0.ɵɵngDeclareInjectable({ minVersion: "12.0.0", version: "12.1.4", ngImport: i0, type: CacheService });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.1.4", ngImport: i0, type: CacheService, decorators: [{
            type: Injectable
        }], ctorParameters: function () { return [{ type: CacheStorageService }]; } });

const CONFIG = new InjectionToken('CONFIG');
let cacheConfigDefaults = {
    keyPrefix: '',
};
function buildCacheService(storage, cacheConfig) {
    cacheConfig = Object.assign(cacheConfigDefaults, cacheConfig);
    return new CacheService(new CacheStorageService(storage, cacheConfig.keyPrefix));
}
class CacheModule {
    static forRoot(cacheConfig) {
        return {
            ngModule: CacheModule,
            providers: [
                { provide: CONFIG, useValue: cacheConfig },
                {
                    provide: CacheService,
                    useFactory: buildCacheService,
                    deps: [Storage, CONFIG],
                },
            ],
        };
    }
}
CacheModule.ɵfac = i0.ɵɵngDeclareFactory({ minVersion: "12.0.0", version: "12.1.4", ngImport: i0, type: CacheModule, deps: [], target: i0.ɵɵFactoryTarget.NgModule });
CacheModule.ɵmod = i0.ɵɵngDeclareNgModule({ minVersion: "12.0.0", version: "12.1.4", ngImport: i0, type: CacheModule, imports: [i1.IonicStorageModule] });
CacheModule.ɵinj = i0.ɵɵngDeclareInjector({ minVersion: "12.0.0", version: "12.1.4", ngImport: i0, type: CacheModule, imports: [[
            IonicStorageModule.forRoot({
                name: '__ionicCache',
                driverOrder: ['indexeddb', 'sqlite', 'websql'],
            }),
        ]] });
i0.ɵɵngDeclareClassMetadata({ minVersion: "12.0.0", version: "12.1.4", ngImport: i0, type: CacheModule, decorators: [{
            type: NgModule,
            args: [{
                    imports: [
                        IonicStorageModule.forRoot({
                            name: '__ionicCache',
                            driverOrder: ['indexeddb', 'sqlite', 'websql'],
                        }),
                    ],
                }]
        }] });

/**
 * Generated bundle index. Do not edit.
 */

export { CONFIG, CacheModule, CacheService, MESSAGES, buildCacheService };
//# sourceMappingURL=ionic-cache.js.map
